#include "pdcontroller.h"
#include "pdcids.h"
#include "vstgui/plugin-bindings/vst3editor.h"
#include "base/source/fstreamer.h"
#include "pluginterfaces/base/ibstream.h"
#include "public.sdk/source/vst/utility/stringconvert.h"
#include "pluginterfaces/base/ustring.h"

#include "base/source/fstreamer.h"
#include "base/source/fstring.h"

using namespace Steinberg;

namespace Steinberg {

class PdParameter : public Vst::Parameter
{
public:
    PdParameter (int32 flags, int32 id);
    void toString (Vst::ParamValue normValue, Vst::String128 string)
            const SMTG_OVERRIDE;
    bool fromString (const Vst::TChar* string, Vst::ParamValue& normValue)
            const SMTG_OVERRIDE;
};

PdParameter::PdParameter(int32 flags, int32 id)
{
    Steinberg::UString (info.title,
            USTRINGSIZE (info.title)).assign (USTRING (" "));
    Steinberg::UString (info.units,
            USTRINGSIZE (info.units)).assign (USTRING (" "));

    info.flags = flags;
    info.id = id;
    info.stepCount = 0;
    info.defaultNormalizedValue = 0;
    info.unitId = Vst::kRootUnitId;
    setNormalized(0);
}

void PdParameter::toString (Vst::ParamValue normValue, Vst::String128 string)
    const
{
    char text[80];
    snprintf(text, 80, "%f", 127. * normValue);
    Steinberg::UString(string, 128).fromAscii(text);
}

bool PdParameter::fromString (const Vst::TChar* string,
    Vst::ParamValue& normValue) const
{
    String wrapper((Vst::TChar*)string);
    double tmp = 0;
    if (wrapper.scanFloat (tmp))
    {
        normValue = tmp / 127.;
        if (normValue < 0)
            normValue = 0;
        else if (normValue > 1)
            normValue = 1;
         return true;
    }
    else return false;
}

tresult PLUGIN_API PdController::initialize (FUnknown* context)
{
    int i;
    tresult result = EditController::initialize (context);
    if (result != kResultOk)
        return result;
    fprintf(stderr, "init controller\n");
    /* create parameters */
    for (i = 0; i < PDNPARAMS; i++) 
    {
        auto *param = new PdParameter(Vst::ParameterInfo::kCanAutomate,
            PdParams::kParam1Id + i);
        parameters.addParameter (param);
    }
    parameters.addParameter (STR16 ("OPEN"), nullptr, 1, 0,
        Vst::ParameterInfo::kCanAutomate, PdParams::kParam1000Id, 0,
            STR16 ("OPEN"));

    return (kResultOk);
}

tresult PLUGIN_API PdController::terminate ()
{
    return EditController::terminate ();
}

tresult PLUGIN_API PdController::setComponentState (IBStream* state)
{
    if (!state)
        return kResultFalse;

    IBStreamer streamer (state, kLittleEndian);

    return kResultOk;  /* not sure why we should do anything here */
}

IPlugView* PLUGIN_API PdController::createView (FIDString name)
{
    if (FIDStringsEqual (name, Vst::ViewType::kEditor))
    {
        auto *view = new VSTGUI::VST3Editor (this, "view", "pdeditor.uidesc");
        Vst::IMessage *msg = allocateMessage();
        msg->setMessageID("open");
        sendMessage(msg);
        msg->release();
        return view;
    }
    return nullptr;
}

tresult PLUGIN_API PdController::setParamNormalized (Vst::ParamID tag,
    Vst::ParamValue value)
{
        /* called by host to update your parameters */
    tresult result = EditController::setParamNormalized (tag, value);
    return result;
}

tresult PLUGIN_API PdController::getParamStringByValue (Vst::ParamID tag, Vst::ParamValue valueNormalized, Vst::String128 string)
{
        /* called by host to get a string for normalized param value */
    if (tag == kParam1000Id)
        return (EditController::getParamStringByValue (tag, valueNormalized,
            string));
    else
    {
        char text[80];
        snprintf(text, 80, "%f", 127. * valueNormalized);
        Steinberg::UString(string, 128).fromAscii(text);
        return (kResultTrue);
    }
}

tresult PLUGIN_API PdController::getParamValueByString (Vst::ParamID tag,
    Vst::TChar* string, Vst::ParamValue& valueNormalized)
{
    /* called by host to get a normalized value from string */
    if (tag == kParam1000Id)
        return (EditController::getParamStringByValue (tag, valueNormalized,
            string));
    else
    {
        String wrapper((Vst::TChar*)string);
        double tmp = 0;
        if (wrapper.scanFloat (tmp))
        {
            valueNormalized = tmp / 127.;
            if (valueNormalized < 0)
                valueNormalized = 0;
            else if (valueNormalized > 1)
                valueNormalized = 1;
        }
        else valueNormalized = 0;
        return (kResultTrue);
    }
}

tresult PLUGIN_API PdController::notify(Vst::IMessage *msg)
{
    uint32 size;
    const void *whatev;
    if (!strcmp(msg->getMessageID(), "quit"))
    {
        fprintf(stderr, "received quit message -- doing nothing\n");
    }
    else if (!strcmp(msg->getMessageID(), "param"))
    {
        if (msg->getAttributes()->getBinary("Text", whatev, size)
            == kResultTrue && size == 80)
        {
            char paramname[80];
            int indx;
            float lolimit, hilimit;
            ((char *)whatev)[79] = 0;   /* isn't this illegal? */
            if (sscanf((char *)whatev, "%d%f%f%s", &indx, &lolimit, &hilimit,
                paramname) == 4 && indx >= 0 && indx < PDNPARAMS)
            {
                Vst::Parameter *p = parameters.getParameterByIndex(indx);
                paramname[15] = 0;  /* limit parameter name length */
                Steinberg::UString (p->getInfo().title,
                    128).fromAscii(paramname);
                Steinberg::UString (p->getInfo().shortTitle,
                    128).fromAscii(paramname);
                if (componentHandler->restartComponent(
                    Vst::kParamTitlesChanged| Vst::kParamValuesChanged |
                        Vst::kReloadComponent) != kResultTrue)
                            fprintf(stderr, "name change failed\n");
            }
            else fprintf(stderr, "bad param message\n");
        }
        else fprintf(stderr, "real bad param message\n");
        return (kResultTrue);   /* lie to host that all's fine anyway */
    }
    else
    {
        fprintf(stderr, "unknown notifier message\n");
        return (kResultFalse);
    }
}


} /* namespace Steinberg */
